from dataclasses import dataclass
from enum import Enum
import re
import networkx as nx

State = str
Interaction = str
TransitionInputType = bool | int


class TransitionType(Enum):
    INPUT = "input"
    OUTPUT = "output"
    INTERNAL = "internal" #implémente les transition de type tau
        
@dataclass
class Variable:
    var_type: TransitionInputType
    var_name: str
    var_domain : list[TransitionInputType]       
    
    def __str__(self):
        return f"Var<{self.var_name} : {self.var_type} : {self.var_domain}>"
    
    def __repr__(self):
        return str(self)

@dataclass
class Transition:
    name: str
    t_type: TransitionType
    guard_condition: str
    from_node: State
    to_node: State
    domains : tuple[Variable]
    
    def __hash__(self):
        return hash((self.name, self.t_type, self.guard_condition, self.from_node, self.to_node))
    
    def __str__(self):
        return f"Trans<{self.name} : {self.t_type} : {self.from_node} -> {self.to_node} : {self.domains}>"

    def __repr__(self):
        return str(self)

@dataclass
class IOSM:
    states: set[State]
    transitions: set[Transition]
    idle_state: State
    
    def to_networkx(self) -> nx.DiGraph:
        G = nx.DiGraph()
        for s in self.states:
            G.add_node(s)
        for t in self.transitions:
            transition_prefix = ""
            match t.t_type:
                case TransitionType.INPUT:
                    transition_prefix = "?"
                case TransitionType.OUTPUT:
                    transition_prefix = "!"
                case TransitionType.INTERNAL:
                    transition_prefix = ""
            
            G.add_edge(t.from_node, t.to_node, label=transition_prefix+t.name)
        return G
   
def find_block(text: str, start_idx: int) -> tuple[str, int]:
    """Retourne le contenu à l'intérieur d'un bloc d'accolades équilibrées,
    ainsi que la position de fin."""
    brace_count = 0
    i = start_idx
    content_start = i
    while i < len(text):
        if text[i] == '{':
            brace_count += 1
            if brace_count == 1:
                content_start = i + 1
        elif text[i] == '}':
            brace_count -= 1
            if brace_count == 0:
                return text[content_start:i], i
        i += 1
    return "", len(text)

def parse_vars(vars_str: str, var_types: dict) -> tuple[Variable, ...]:
    '''Analyse la liste de variables dans les parenthèses, ex: ( c, x )'''
    if not vars_str.strip():
        return ()
    items = [v.strip() for v in vars_str.split(',')]
    
    return tuple(
        Variable(
            var_type=var_types.get(item, bool),
            var_name=item,
            var_domain=[]
        )
        for item in items
    )

def IOSM_from_XLIA(filepath : str) -> IOSM:
    with open(filepath, "r") as f:
        lines = f.read()
    states = set()
    transitions = set()
    idle_state = None
    var_types = {}

    # Extraction des variables système du bloc @private
    var_pattern = re.compile(r'var\s+([\w\s]+)\s+(\w+)\s*;')
    for var_match in var_pattern.finditer(lines):
        var_type_str = var_match.group(1).strip()
        var_name = var_match.group(2).strip()
        # Déduction du type Python
        if 'urational' in var_type_str:
            var_type = float
        elif 'pos_integer' in var_type_str:
            var_type = int  
        else :
            var_type = bool
        var_types[var_name] = var_type
    
    # Modifié pour mieux détecter l'état initial
    pattern = re.compile(r'state\s*(?:<\s*start\s*>)?\s*(\w+)\s*\{', re.DOTALL)
    for match in pattern.finditer(lines):
        current_state = match.group(1)
        states.add(current_state)
        if "<start>" in match.group(0) or "<initial>" in match.group(0):
            idle_state = current_state
        
        block_content, end_pos = find_block(lines, match.end() - 1)
        
        # Modifié pour gérer les transitions simples
        transition_pattern = r'transition\s+(\w+)\s*-->\s*(\w+)\s*\{([^}]*)\}'
        for tmatch in re.finditer(transition_pattern, block_content, re.DOTALL):
            t_name = tmatch.group(1)
            t_to   = tmatch.group(2)
            body   = tmatch.group(3).strip()

            t_type = TransitionType.INTERNAL
            domains = ()
            
            # Nouveau: chercher output/input n'importe où dans le corps de la transition
            out_match = re.search(r'output\s+(\w+)\s*(?:\(([^)]*)\))?', body)
            in_match = re.search(r'input\s+(\w+)\s*(?:\(([^)]*)\))?', body)
            
            if in_match:
                t_type = TransitionType.INPUT
                vars_str = in_match.group(2) if in_match.group(2) else ""
                domains = parse_vars(vars_str, var_types)
            elif out_match:
                t_type = TransitionType.OUTPUT
                vars_str = out_match.group(2) if out_match.group(2) else ""
                domains = parse_vars(vars_str, var_types)

            g_match = re.search(r'guard\s*\(\s*([^)]+)\s*\)', body)
            tg_match = re.search(r'tguard\s*\(\s*([^)]+)\s*\)', body)
            
            guard_cond = g_match.group(1) if g_match else ""
            if tg_match:
                guard_cond += (" && " + tg_match.group(1)) if guard_cond else tg_match.group(1)

            transitions.add(
                Transition(
                    name=t_name,
                    t_type=t_type,
                    guard_condition=guard_cond,
                    from_node=current_state,
                    to_node=t_to,
                    domains=domains
                )
            )
    
    return IOSM(states=states, transitions=transitions, idle_state=idle_state)