from oracle import Oracle
from multigraph import BipartiteMultiGraph
from json_to_x import json_to_x
import argparse
import subprocess as sp
from enum import Enum
from dataclasses import dataclass
import time


def get_args():
    parser = argparse.ArgumentParser(description='ATM Test Generator')
    parser.add_argument('--xlia_fp', type=str, required=True, help='Path to XLIA file')
    parser.add_argument('--json_fp', type=str, required=True, help='Path to JSON test cases')
    parser.add_argument('--bb_fp', type=str, required=True, help='Path to black box simulator file')
    parser.add_argument('--max_nodes', type=int, default=100, help='Maximum number of nodes to visit')
    args = parser.parse_args()
    return args


class SUTEventType(Enum):
    INPUT = "bb_input"
    OUTPUT = "bb_output"
    NEW_STATE = "bb_newstate"

@dataclass
class SUTEvent:
    type : SUTEventType
    content : str

class BBHandler:
    
    def __init__(self, bb_fp : str) -> None:
        self.bb_fp = bb_fp
        process = sp.Popen(
            ["python", bb_fp],  
            stdin=sp.PIPE,
            stdout=sp.PIPE,
            stderr=sp.PIPE,
            text=True
        )
        self.process = process
        self.is_waiting_for_input = False
        self.state = 'Idle' 
        
    def send_msg(self, message) -> None:
        self.process.stdin.write(f"{message}\n")
        self.process.stdin.flush()
    
    def read(self) -> SUTEvent:
        """renvoie un SUTEvent"""
        line = next(self.process.stdout).strip()
        if "in_" in line:
            self.is_waiting_for_input = True
            return SUTEvent(
                type=SUTEventType.INPUT,
                content=line.split("_", 1)[1] # variable demandée
            )
        elif "out_" in line:
            self.is_waiting_for_input = False
            return SUTEvent(
                type=SUTEventType.OUTPUT,
                content=line.split("_", 1)[1]
            )
        elif "state_" in line:
            self.is_waiting_for_input = False
            self.state= line.split("_", 1)[1]
            return SUTEvent(
                type=SUTEventType.NEW_STATE,
                content=self.state
            )
    
    def close(self) -> None:
        self.process.terminate()
    

def verdict(xlia_fp:str, json_fp : str, bb_fp : str, max_nodes: int, nb_steps=0) -> None:
    g = BipartiteMultiGraph(xlia_fp)
    x = json_to_x(json_fp)
    print(x)
    oracle = Oracle(max_nodes, x, g)
    oracle.current_node_visited = nb_steps# se réferer au try: except: dans la boucle while
    
    black_box_handler = BBHandler(bb_fp)

    def check_exit():
        if oracle.check_fail():
            print("FAIL")
            black_box_handler.close()
            exit(0)
        elif oracle.check_pass():
            print("PASS")
            black_box_handler.close()
            exit(0)
        elif oracle.MAX_NODE_VISITED <= oracle.current_node_visited:
            print("INCONCLUSIVE: MAX NODES VISITED")
            black_box_handler.close()
            exit(0)

    emitted_variables = {}
    while True:
        #first check if test case is done
        
        check_exit()
    
        # the oracle plays first
        event = black_box_handler.read()
        match event.type:
            case SUTEventType.INPUT:
                #send the variable
                try:
                    # une erreur à ce stade signifie que l'input aléatoire que notre testeur a envoyé 
                    # n'a pas permis de réaliser la transition souhaitée, on fait le choix
                    black_box_handler.send_msg(emitted_variables[event.content])
                except:
                    print("[partial fail due to random inputs]: initializing test with different set of random inputs (retrying)")
                    return verdict(xlia_fp, json_fp, bb_fp, max_nodes, nb_steps = oracle.current_node_visited)
                    
            case SUTEventType.OUTPUT:
                print(f"SUT Output: {event.content}")
            case SUTEventType.NEW_STATE:
                oracle.update_move(event.content)
                transi, new_node, emitted_variables = oracle.next_move()
                print ("Tester Input :\n" + transi + f" --{emitted_variables}--> " + new_node)
                black_box_handler.send_msg(transi)
                pass
            
        #maybe his move eneded the game
        check_exit()
        
        

if __name__ == "__main__":
    args = get_args()
    verdict(args.xlia_fp, args.json_fp, args.bb_fp, args.max_nodes)
