from dataclasses import dataclass
from IOSM import State, IOSM_from_XLIA, IOSM, TransitionType, Transition
import networkx as nx
import sys
import matplotlib.pyplot as plt

label = str

@dataclass
class Graph:
    """Graph that represents an accessibility game"""
    Vb: list[State]
    Vr: list[State]
    E : dict[State, list[State]]
    E_label : dict[tuple[State, State], Transition]
    idle: State
    
    def render_graph(self)->None:
        """Render the graph using networkx and matplotlib"""
        G = nx.DiGraph()
        G.add_nodes_from(self.Vb, bipartite=0)
        G.add_nodes_from(self.Vr, bipartite=1)
        for s, targets in self.E.items():
            for target in targets:
                t = self.E_label[(s, target)]
                transition_prefix = ""
                match t.t_type:
                    case TransitionType.INPUT:
                        transition_prefix = "?"
                    case TransitionType.OUTPUT:
                        transition_prefix = "!"
                    case TransitionType.INTERNAL:
                        transition_prefix = ""
                G.add_edge(s, target, label=transition_prefix+self.E_label[(s, target)].name,)
        
        color_map = []
        for node in G.nodes():
            if node == self.idle:
                color_map.append("green")
            else:
                color_map.append("blue" if node in self.Vb else "red")
        
        pos = nx.spring_layout(G,k=30, seed=42)
        
        nx.draw_networkx_nodes(G, pos, node_color=color_map)
        nx.draw_networkx_labels(G, pos)
        nx.draw_networkx_edges(
            G,
            pos,
            edge_color="gray",
            arrows=True,
        )
        edge_labels = {(s, t): data['label'] for s, t, data in G.edges(data=True)}
        nx.draw_networkx_edge_labels(G, pos, edge_labels=edge_labels, label_pos=0.8)
        if not nx.algorithms.is_bipartite(G):
            print("[Graph render WARNING]: graph is not bipartite")
        plt.show()
    
    
BOT = State("BOT")

class IOSMGraph(Graph):
    """Graph associated to an IOSM specification"""
    def __init__(self, iosm : IOSM) -> None:
        Vr_set = {
                state for state in iosm.states
                if all(t.t_type == TransitionType.OUTPUT for t in iosm.transitions if t.from_node == state)
            }
        Vr = Vr_set
        Vb: set[State] = (iosm.states - Vr_set).union("⊥")  # Include sink node in Vb

        # Initialize edges and labels
        E: dict[str, list[str]] = {state: [] for state in iosm.states}
        E["⊥"] = []  # Ensure sink node is in E
        E_label: dict[tuple[str, str], Transition] = {}
        intermediate_counter = 0

        # Add edges from Vr states to ⊥ with dummy transitions
        

        # Track intermediates and update Vr/Vb dynamically

        # Process each IOSM transition
        for trans in iosm.transitions:
            src, dest = trans.from_node, trans.to_node
            src_in_vr = src in Vr_set
            dest_in_vr = dest in Vr_set

            if src_in_vr != dest_in_vr:
                # Direct edge between different partitions
                E[src].append(dest)
                E_label[(src, dest)] = trans
                if trans.t_type == TransitionType.INTERNAL:
                    print("[IOSMGraph.__init__ WARNING]: original iosm transition is internal")
            else:
                # Insert intermediate node in opposite partition
                intermediate = f"inter_{intermediate_counter}"
                intermediate_counter += 1
                if src_in_vr:
                    # Intermediate in Vb
                    Vb.add(intermediate)
                    Vr.add(dest)
                    Vb.discard(dest)
                else:
                    # Intermediate in Vr
                    Vr.add(intermediate)
                    Vr.discard(dest)
                    Vb.add(dest)

                # Split edge into src -> intermediate -> dest
                E[src].append(intermediate)
                E_label[(src, intermediate)] = trans

                E[intermediate] = [dest]
                E_label[(intermediate, dest)] = Transition( 
                                                           name="tau",
                                                              t_type=TransitionType.INTERNAL,
                                                              guard_condition="true",
                                                              from_node=intermediate,
                                                              to_node=dest,
                                                              domains=()
                )
        # Add dummy transitions to sink node
        for state in Vr:
            dummy_trans = Transition(
                name="to_bot",
                t_type=TransitionType.INTERNAL,
                guard_condition="",
                from_node=state,
                to_node="⊥",
                domains=()
            )
            edge = (state, "⊥")
            E[state].append("⊥")
            E_label[edge] = dummy_trans
        #ajouter un noeud liée au sink pour éviter des deadlocks
        
        intermediate = "inter_⊥"
        E[intermediate] = ["⊥"]
        E["⊥"].append(intermediate)
        E_label[(intermediate, "⊥")] = Transition( 
                                                           name="tau",
                                                              t_type=TransitionType.INTERNAL,
                                                              guard_condition="true",
                                                              from_node=intermediate,
                                                              to_node="⊥",
                                                              domains=()
                )
        E_label[("⊥", intermediate)] = Transition( 
                                                           name="tau",
                                                              t_type=TransitionType.INTERNAL,
                                                              guard_condition="true",
                                                              from_node="⊥",
                                                              to_node=intermediate,
                                                              domains=()
                )
         
        
        # Remove duplicates and sort for consistency
        Vb = sorted(list(Vb))
        Vr = sorted(list(Vr))

        super().__init__( 
                         Vb, 
                         Vr, 
                         {k: sorted(v) for k, v in E.items() if v}, # Clean empty entries
                         E_label, 
                         iosm.idle_state
                         )

                    
    
                
        
        
if __name__ == "__main__":
    if len(sys.argv) < 2:
        print("Usage: graph.py <filepath>")
        sys.exit(1)

    filepath = sys.argv[1]
    iosm = IOSM_from_XLIA(filepath+".xlia")
    print("\n".join(map(str, iosm.transitions)))    
    g = IOSMGraph(iosm)
    g.render_graph()