from IOSM import *
import networkx as nx
import matplotlib.pyplot as plt

class MultiGraph:
    def __init__(self, Vr: list, Vb: list, E: dict, E_label: dict, idle: str):
        self.Vr = Vr
        self.Vb = Vb
        self.n = len(Vb) + len(Vr)
        self.E = E
        self.E_label = E_label
        self.idle = idle

class IOSMMultiGraph(MultiGraph):
    def __init__(self, XLIApath: str):
        iosm = IOSM_from_XLIA(XLIApath)
        states = iosm.states
        Vr, Vb = set(), set()
        E = {state: [] for state in states}
        E_transit = {state: {} for state in states}

        E_label = {}
        for transit in iosm.transitions:
            if(transit.name !="Timeout"):
                E[transit.from_node].append(transit.to_node)
                label = ("?" if transit.t_type == TransitionType.INPUT else 
                        ("!" if transit.t_type == TransitionType.OUTPUT else "")) + transit.name
                
                if (transit.from_node, transit.to_node) not in E_label:
                    E_label[transit.from_node, transit.to_node] = []
                E_label[transit.from_node, transit.to_node].append(label)

                if transit.to_node not in E_transit[transit.from_node]:
                    E_transit[transit.from_node][transit.to_node] = []
                E_transit[transit.from_node][transit.to_node].append(transit.t_type)

        for u in states:
            for v, types in E_transit[u].items():
                if not all(t == TransitionType.OUTPUT for t in types):
                    Vb.add(u)
                    break
            if u not in Vb:
                Vr.add(u)
        self.transis = iosm.transitions
        self.variables_output ={}
        for t in self.transis:
            self.variables_output[t.from_node, t.to_node] = t.domains
                 
        self.iosm = iosm

    
        super().__init__(list(Vr), list(Vb), E, E_label, iosm.idle_state)

class BipartiteMultiGraph(IOSMMultiGraph):
    def __init__(self, XLIApath: str):
        super().__init__(XLIApath)
        #self.variables_output = self.iosm.transitions

        Bip_Vr = self.Vr.copy()
        Bip_Vb = self.Vb.copy()
        Bip_E = {k: v.copy() for k, v in self.E.items()}
        Bip_E_label = {(k1, k2): v.copy() for (k1, k2), v in self.E_label.items()}

        bottom_node = "bot"
        Bip_Vb.append(bottom_node)
        
        error_node = "err"
        Bip_Vr.append(error_node)

        Bip_E[bottom_node] = [error_node]
        Bip_E_label[bottom_node, error_node] = ["tau_" + error_node]

        Bip_E[error_node] = [bottom_node]
        Bip_E_label[error_node, bottom_node] = ["tau_" + bottom_node]
        
        tau_counter = 0
        for u in self.Vb:
            for v in self.E[u]:
                if v in self.Vb:
                    labels = Bip_E_label.get((u, v), []).copy() 
                    for label in labels:
                        tau_node = f"tau_{tau_counter}"
                        tau_counter += 1
                        Bip_Vr.append(tau_node)

                        Bip_E[u].append(tau_node)
                        Bip_E[u].remove(v)

                        if tau_node not in Bip_E:
                            Bip_E[tau_node] = [v]
                        else:
                            Bip_E[tau_node].append(v)

                        Bip_E[tau_node].append(bottom_node)
                        Bip_E_label[tau_node, bottom_node] = ["tau_" + bottom_node]

                        Bip_E_label[u, tau_node] = [label]
                        Bip_E_label[tau_node, v] = ["tau_" + v]

                        if  (u, v) in self.variables_output:
                            self.variables_output[u, tau_node] = self.variables_output[u, v]
                            self.variables_output[tau_node, u] = set()

                            del self.variables_output[u, v]
                        if (u, v) in Bip_E_label:
                            Bip_E_label[u, v].remove(label)

        for u in self.Vr:
            for v in self.E[u]:
                if v in self.Vr:
                    labels = Bip_E_label.get((u, v), []).copy() 
                    for label in labels:
                        tau_node = f"tau_{tau_counter}"
                        tau_counter += 1
                        Bip_Vb.append(tau_node)
                        Bip_E[u].append(tau_node)
                        Bip_E[u].remove(v)
                        
                        if tau_node not in Bip_E:
                            Bip_E[tau_node] = [v]
                        else:
                            Bip_E[tau_node].append(v)

                        Bip_E_label[u, tau_node] = [label]
                        Bip_E_label[tau_node, v] = ["tau_" + v]
                        if (u, v) in Bip_E_label:
                            Bip_E_label[u, v].remove(label)
                        if(u, v) in self.variables_output:
                            self.variables_output[u, tau_node] = self.variables_output[u, v]
                            self.variables_output[tau_node, u] = set()
                            del self.variables_output[u, v]           
            Bip_E[u].append(bottom_node) 
            Bip_E_label[u, bottom_node] = ["tau_" + bottom_node]
        
        self.Vr = Bip_Vr
        self.Vb = Bip_Vb
        self.E = Bip_E
        self.E_label = Bip_E_label

        G = nx.DiGraph()
        G.add_nodes_from(self.Vb, bipartite=0)
        G.add_nodes_from(self.Vr, bipartite=1)
        for s, targets in self.E.items():
            for target in targets:
                G.add_edge(s, target, label=self.E_label[s, target])

        self.G = G

    def render_graph(self)->None:
        """Render the graph using networkx and matplotlib"""
        G = self.G
        color_map = []
        for node in G.nodes():
            if node == self.idle:
                color_map.append("green")
            else:
                color_map.append("blue" if node in self.Vb else "red")
            
        pos = nx.spring_layout(G,k=30, seed=42)
            
        nx.draw_networkx_nodes(G, pos, node_color=color_map)
        nx.draw_networkx_labels(G, pos)
        nx.draw_networkx_edges(
            G,
            pos,
            edge_color="gray",
            arrows=True,
        )
        edge_labels = {(s, t): data['label'] for s, t, data in G.edges(data=True)}
        nx.draw_networkx_edge_labels(G, pos, edge_labels=edge_labels, label_pos=0.8)
        plt.show(block=False)
        plt.pause(0.001)